from eth_rpc import ContractFunc, ProtocolBase
from eth_rpc.types import METHOD, NoArgs, primitives


class ProtocolRegistry(ProtocolBase):
    blocks: ContractFunc[
        NoArgs,
        primitives.address,
    ] = METHOD

    blockstorage: ContractFunc[
        NoArgs,
        primitives.address,
    ] = METHOD



from decimal import Decimal
from typing import Annotated

from eth_rpc.contract import ContractFunc
from eth_rpc.types import METHOD, Name, NoArgs, primitives
from eth_typeshed.erc20 import ERC20
from eth_typing import HexAddress
from pydantic import BaseModel


class GetReservesResponse(BaseModel):
    reserve0: Annotated[primitives.uint112, Name("_reserve0")]
    reserve1: Annotated[primitives.uint112, Name("_reserve1")]
    timestamp: Annotated[primitives.uint32, Name("_blockTimestampLast")]


class UniswapV2Pair(ERC20):
    get_reserves: Annotated[
        ContractFunc[
            NoArgs,
            GetReservesResponse,
        ],
        Name("getReserves"),
    ] = METHOD
    token0: ContractFunc[NoArgs, HexAddress] = METHOD
    token1: ContractFunc[NoArgs, HexAddress] = METHOD

    def get_price(self, token0: bool = True):
        reserves0, reserves1, _ = self.get_reserves().get()
        price0, price1 = (reserves1 / reserves0, reserves0 / reserves1)
        return Decimal(price0) if token0 else Decimal(price1)



from enum import IntEnum

from eth_rpc.types import Bytes32Hex as Bytes32
from eth_rpc.types import primitives
from eth_rpc.types.struct import Struct


class CommandStatus(IntEnum):
    PENDING = 0
    SUCCESS = 1
    FAILURE = 2
    GAS_USED = 3


class Reference(Struct):
    platform: Bytes32
    identifier: Bytes32


class Field(Struct):
    key: Bytes32
    value: primitives.string


class Command(Struct):
    id: Bytes32 = primitives.bytes32(b"")
    chain_id: primitives.uint256 = primitives.uint256(0)
    nonce: primitives.uint256 = primitives.uint256(0)
    value: primitives.uint256 = primitives.uint256(0)
    owner_id: Bytes32
    namespace: Bytes32 = primitives.bytes32(b"")
    wallet_index: primitives.uint16 = primitives.uint16(0)

    metaprotocol: Bytes32
    modifiers: list[Bytes32] = []
    references: list[Reference] = []
    args: list[primitives.string] = []
    kwargs: list[Field] = []

    timestamp: primitives.uint64 = primitives.uint64(0)
    gas: primitives.uint256 = primitives.uint256(0)
    bribe: primitives.uint256 = primitives.uint256(0)
    hash: Bytes32 = primitives.bytes32(b"")
    status: CommandStatus = CommandStatus.PENDING
    gas_used: primitives.uint256 = primitives.uint256(0)

    def __repr__(self):
        protocol = self.metaprotocol.decode("utf-8").rstrip("\x00")
        namespace = self.namespace.decode("utf-8").rstrip("\x00")
        return (
            f'<Command namespace="{namespace}" protocol="{protocol}" args={self.args}>'
        )

    __str__ = __repr__



from typing import Annotated

from eth_rpc.event import Event
from eth_rpc.types import Indexed, primitives
from eth_typing import HexAddress
from pydantic import BaseModel


class StakeType(BaseModel):
    executor: Annotated[HexAddress, Indexed]
    user: Annotated[HexAddress, Indexed]
    amount: primitives.uint256


class UnstakeType(BaseModel):
    user: Annotated[HexAddress, Indexed]
    amount: primitives.uint256


StakeEvent = Event[StakeType](name="Stake")
UnstakeEvent = Event[UnstakeType](name="Unstake")



from .events.index_manager import (
    AddIndexEvent,
    OwnershipTransferredEvent,
    RemoveIndexEvent,
    SetVerifiedEvent,
)

# TODO: investigate, is this needed?
from .events.peas_protocol_fee import SetYieldAdminEvent, SetYieldBurnEvent
from .events.staking_pool import StakeEvent, UnstakeEvent
from .events.token_rewards import (
    AddSharesEvent,
    ClaimRewardEvent,
    DepositRewardsEvent,
    DistributeRewardEvent,
    RemoveSharesEvent,
)
from .events.weighted_index import (
    AddLiquidityEvent,
    BondEvent,
    CreateEvent,
    DebondEvent,
    FlashLoanEvent,
    RemoveLiquidityEvent,
)
from .index_manager import IndexManager
from .staking_pool import StakingPool
from .token_rewards import TokenRewards
from .weighted_index import WeightedIndex

__all__ = [
    "IndexManager",
    "StakingPool",
    "WeightedIndex",
    "TokenRewards",
    "AddIndexEvent",
    "RemoveIndexEvent",
    "SetVerifiedEvent",
    "OwnershipTransferredEvent",
    "SetYieldAdminEvent",
    "SetYieldBurnEvent",
    "AddSharesEvent",
    "RemoveSharesEvent",
    "ClaimRewardEvent",
    "DistributeRewardEvent",
    "DepositRewardsEvent",
    "StakeEvent",
    "UnstakeEvent",
    "CreateEvent",
    "FlashLoanEvent",
    "AddLiquidityEvent",
    "RemoveLiquidityEvent",
    "BondEvent",
    "DebondEvent",
]



from typing import Annotated

from eth_rpc import ContractFunc, ProtocolBase
from eth_rpc.types import METHOD, Name
from eth_typing import HexAddress, HexStr
from pydantic import BaseModel

V2_FACTORY_ADDRESS = HexAddress(HexStr("0x5C69BEE701EF814A2B6A3EDD4B1652CB9CC5AA6F"))


class GetPairRequest(BaseModel):
    token_a: Annotated[HexAddress, Name("tokenA")]
    token_b: Annotated[HexAddress, Name("tokenB")]


class UniswapV2Factory(ProtocolBase):
    get_pair: Annotated[ContractFunc[GetPairRequest, HexAddress], Name("getPair")] = (
        METHOD
    )



from .base_source import BaseSource, SimulacrumSubmission
from .command import Command, CommandStatus
from .twitter import Tweet

__all__ = [
    "BaseSource",
    "Command",
    "CommandStatus",
    "SimulacrumSubmission",
    "Tweet",
]



from typing import Annotated

from eth_rpc import ContractFunc, ProtocolBase
from eth_rpc.types import METHOD, Name, NoArgs, primitives


class Mempool(ProtocolBase):
    size_of: Annotated[
        ContractFunc[
            NoArgs,
            primitives.uint256,
        ],
        Name("sizeOf"),
    ] = METHOD



from typing import Annotated, NamedTuple  # noqa: D100

from eth_rpc.contract import ContractFunc
from eth_rpc.types import METHOD, Name, NoArgs, primitives
from eth_typeshed.erc20 import ERC20
from eth_typing import HexAddress
from pydantic import BaseModel


class IndexAssetInfo(NamedTuple):
    token: HexAddress
    weighting: primitives.uint256
    basePriceUSDX96: primitives.uint256
    c1: HexAddress
    q1: primitives.uint256


class IndexAssets(BaseModel):
    values: list[IndexAssetInfo]


class Fees(BaseModel):
    burn: primitives.uint256
    bond: primitives.uint256
    debond: primitives.uint256
    buy: primitives.uint256
    sell: primitives.uint256
    partner: primitives.uint256


class WeightedIndex(ERC20):
    get_all_assets: Annotated[
        ContractFunc[NoArgs, IndexAssets], Name("getAllAssets")
    ] = METHOD
    get_idx_price_USDX96: Annotated[
        ContractFunc[NoArgs, tuple[primitives.uint256, primitives.uint256]],
        Name("getIdxPriceUSDX96"),
    ] = METHOD
    index_tokens: Annotated[
        ContractFunc[primitives.uint256, IndexAssetInfo], Name("indexTokens")
    ] = METHOD
    indextype: Annotated[ContractFunc[NoArgs, primitives.uint8], Name("indexType")] = (
        METHOD
    )
    bond_fee: Annotated[ContractFunc[NoArgs, primitives.uint256], Name("BOND_FEE")] = (
        METHOD
    )
    debond_fee: Annotated[
        ContractFunc[NoArgs, primitives.uint256], Name("DEBOND_FEE")
    ] = METHOD
    is_asset: Annotated[ContractFunc[HexAddress, primitives.bool], Name("isAsset")] = (
        METHOD
    )
    lp_rewards_token: Annotated[
        ContractFunc[NoArgs, HexAddress], Name("lpRewardsToken")
    ] = METHOD
    lp_staking_pool: Annotated[
        ContractFunc[NoArgs, HexAddress], Name("lpStakingPool")
    ] = METHOD
    paired_lp_token: Annotated[
        ContractFunc[NoArgs, HexAddress], Name("PAIRED_LP_TOKEN")
    ] = METHOD
    partner: ContractFunc[NoArgs, HexAddress] = METHOD
    fees: ContractFunc[NoArgs, Fees] = METHOD
    created: ContractFunc[NoArgs, primitives.uint256] = METHOD
    index_type: Annotated[ContractFunc[NoArgs, primitives.uint8], Name("indexType")] = (
        METHOD
    )



from typing import Annotated, ClassVar

from eth_rpc import ContractFunc
from eth_rpc.types import METHOD, Name, NoArgs, primitives

from ..utils import BaseSource, Command


class CommandSource(BaseSource):
    source_name: ClassVar[str] = "Command"
    command_type: ClassVar[type] = Command

    callback_address: Annotated[
        ContractFunc[
            NoArgs,
            primitives.address,
        ],
        Name("callbackAddress"),
    ] = METHOD

    lookup_address: Annotated[
        ContractFunc[
            primitives.address,
            primitives.address,
        ],
        Name("lookupAddress"),
    ]



from typing import Annotated

from eth_rpc import ContractFunc
from eth_rpc.contract.base import ProtocolBase
from eth_rpc.types import NoArgs, primitives


class ValidatorRegistry(ProtocolBase):
    threshold: ContractFunc[
        NoArgs,
        Annotated[primitives.uint8, "count"],
    ]

    validators: ContractFunc[
        primitives.address,
        bool,
    ]



from .events import (
    V3PoolCreatedEvent,
    V3PoolCreatedEventType,
    V3SwapEvent,
    V3SwapEventType,
)
from .factory import GetPoolRequest, UniswapV3Factory
from .nonfungible_position_manager import (
    NONFUNGIBLE_POSITION_MANAGER_ADDRESS,
    NonfungiblePositionManager,
)
from .pool import ProcessedTick, Slot0, Tick, UniswapV3Pool

__all__ = [
    "NONFUNGIBLE_POSITION_MANAGER_ADDRESS",
    "GetPoolRequest",
    "NonfungiblePositionManager",
    "ProcessedTick",
    "Slot0",
    "Tick",
    "UniswapV3Factory",
    "UniswapV3Pool",
    "V3PoolCreatedEvent",
    "V3PoolCreatedEventType",
    "V3SwapEvent",
    "V3SwapEventType",
]



import asyncio
import math
from collections.abc import Iterable
from decimal import Decimal
from typing import Annotated, cast

from eth_rpc import ContractFunc, ProtocolBase
from eth_rpc.types import METHOD, Name, NoArgs, primitives
from eth_typeshed.multicall import multicall
from eth_typing import HexAddress
from pydantic import BaseModel, Field, PrivateAttr

from ..erc20 import ERC20
from .constants import MIN_TICK, Q192
from .events import V3MintEvent, V3SwapEvent, V3SwapEventType
from .utils import tick_from_bitmap, tick_to_price


class Slot0(BaseModel):
    sqrt_price: primitives.uint160 = Field(serialization_alias="sqrtPriceX96")
    tick: primitives.int24
    observation_index: primitives.uint16 = Field(serialization_alias="observationIndex")
    observation_cardinality: primitives.uint16 = Field(
        serialization_alias="observationCardinality"
    )
    observation_cardinality_next: primitives.uint16 = Field(
        serialization_alias="observationCardinality"
    )
    fee_protocol: primitives.uint8 = Field(serialization_alias="feeProtocol")
    unlocked: bool


class Tick(BaseModel):
    liquidity_gross: Annotated[primitives.uint128, Name("liquidityGross")]
    liquidity_net: Annotated[primitives.int128, Name("liquidityNet")]
    fee_growth_outside0: Annotated[primitives.uint256, Name("feeGrowthOutside0X128")]
    fee_growth_outside1: Annotated[primitives.uint256, Name("feeGrowthOutside1X128")]
    tick_cumulative_outside: Annotated[primitives.int56, Name("tickCumulativeOutside")]
    seconds_per_liquidity_outside: Annotated[
        primitives.uint160, Name("secondsPerLiquidityOutsideX128")
    ]
    seconds_outside: Annotated[primitives.uint32, Name("secondsOutside")]
    initialized: bool  # liquidity_net > 0

    def __repr__(self):
        return f"<Tick liquidity_net={self.liquidity_net}>"


class ProcessedTick(BaseModel):
    index: int
    tick: Tick
    token0: ERC20
    token1: ERC20
    active_liquidity: int = 0

    @property
    def liquidity_net(self):
        return self.tick.liquidity_net

    @property
    def price0(self):
        return tick_to_price(
            self.index, self.token0.get_decimals(), self.token1.get_decimals()
        )

    @property
    def price1(self):
        return (
            tick_to_price(
                self.index, self.token0.get_decimals(), self.token1.get_decimals()
            )
            ** -1
        )

    def __repr__(self):
        return f"<ProcessedTick index={self.index}>"


class UniswapV3Pool(ProtocolBase):
    _tick_spacing: int | None = PrivateAttr(None)
    _token0: ERC20 | None = PrivateAttr(None)
    _token1: ERC20 | None = PrivateAttr(None)

    tick_bitmap: Annotated[
        ContractFunc[primitives.int16, primitives.uint256],
        Name("tickBitmap"),
    ] = METHOD
    fee: ContractFunc[NoArgs, primitives.uint24] = METHOD
    slot0: ContractFunc[NoArgs, Slot0] = METHOD
    token0: ContractFunc[NoArgs, HexAddress] = METHOD
    token1: ContractFunc[NoArgs, HexAddress] = METHOD
    ticks: ContractFunc[primitives.int24, Tick] = METHOD
    liquidity: ContractFunc[NoArgs, primitives.uint128] = METHOD
    tick_spacing: Annotated[
        ContractFunc[NoArgs, primitives.int24],
        Name("tickSpacing"),
    ] = METHOD
    fee_growth_global0: Annotated[
        ContractFunc[NoArgs, primitives.uint256], Name("feeGrowthGlobal0X128")
    ] = METHOD
    fee_growth_global1: Annotated[
        ContractFunc[NoArgs, primitives.uint256], Name("feeGrowthGlobal1X128")
    ] = METHOD

    async def get_price(self, token0: bool = True):
        slot0 = await self.slot0().get()
        _token0 = await self.token0().get()
        _token1 = await self.token1().get()
        token0_decimals = await ERC20(address=_token0).decimals().get()
        token1_decimals = await ERC20(address=_token1).decimals().get()
        return self.sqrt_price_x96_to_token_prices(
            slot0.sqrt_price,
            token0_decimals,
            token1_decimals,
            token0,
        )

    @staticmethod
    def sqrt_price_x96_to_token_prices(
        sqrt_price_x96: int,
        token0_decimals: int,
        token1_decimals: int,
        token0: bool = True,
    ) -> Decimal:
        num = Decimal(sqrt_price_x96 * sqrt_price_x96)
        price1 = (
            (num / Q192)
            * (Decimal(10) ** token0_decimals)
            / (Decimal(10) ** token1_decimals)
        )
        price0 = Decimal(1) / price1

        return price0 if token0 else price1

    async def all_tick_values(self):
        tick_spacing = await self.get_tick_spacing()
        min_tick, max_tick = await self.tick_range()
        return list(range(min_tick, max_tick, tick_spacing))

    async def get_token0(self):
        if not self._token0:
            self._token0 = ERC20(await self.token0().get())
        return self._token0

    async def get_token1(self):
        if not self._token1:
            self._token1 = ERC20(await self.token1().get())
        return self._token1

    async def get_tick_spacing(self):
        if not self._tick_spacing:
            self._tick_spacing = await self.tick_spacing().get()
        return self._tick_spacing

    async def get_swaps(
        self, start_block, end_block
    ) -> dict[int, list[V3SwapEventType]]:
        ticks: dict[int, list[V3SwapEventType]] = {}
        tick_values = await self.all_tick_values()
        async for event in V3SwapEvent.set_filter(addresses=[self.address]).backfill(
            start_block, end_block
        ):
            tick = event.event.tick
            closest_tick = max(t for t in tick_values if t <= tick)
            if tick in ticks:
                ticks[closest_tick].append(event.event)
            else:
                ticks[closest_tick] = [event.event]
        return ticks

    async def get_ticks(self, indices: Iterable[int]) -> list[ProcessedTick]:
        calls = []
        for tick in indices:
            calls.append(self.ticks(primitives.int24(tick)))
        ticks: list[Tick] = await multicall.execute(*calls)

        processed_ticks = []
        for idx, tick_data in zip(indices, ticks):
            processed_ticks.append(
                ProcessedTick(
                    index=idx,
                    tick=tick_data,
                    token0=await self.get_token0(),
                    token1=await self.get_token1(),
                )
            )
        return processed_ticks

    async def tvl(self, block_number=None) -> list[ProcessedTick]:
        liquidity, initialized_ticks, active_tick_index = await asyncio.gather(
            self.liquidity().get(),
            self.get_initialized_ticks(),
            self.active_tick(),
        )
        ticks = await self.get_ticks(initialized_ticks)

        if active_tick_index in initialized_ticks:
            active_tick = [t for t in ticks if t.index == active_tick_index][0]
            active_tick.active_liquidity = liquidity
            current_liquidity = liquidity - active_tick.liquidity_net
            previous_tick_net = active_tick.liquidity_net
        else:
            current_liquidity = liquidity
            previous_tick_net = 0

        for tick in [tick for tick in ticks if tick.index < active_tick_index][::-1]:
            tick.active_liquidity = current_liquidity - previous_tick_net
            current_liquidity = tick.active_liquidity
            previous_tick_net = tick.liquidity_net

        current_liquidity = liquidity
        previous_tick_net = 0
        for tick in [tick for tick in ticks if tick.index > active_tick_index]:
            tick.active_liquidity = current_liquidity + tick.liquidity_net
            current_liquidity = tick.active_liquidity
        return ticks

    async def get_tick_range(self, lower: int, upper: int) -> list[ProcessedTick]:
        if not self._tick_spacing:
            self._tick_spacing = await self.tick_spacing().get()
        assert self._tick_spacing
        return await self.get_ticks(range(lower, upper, self._tick_spacing))

    async def tick_range(self):
        tick_spacing = await self.get_tick_spacing()
        min_tick = math.ceil(MIN_TICK / tick_spacing) * tick_spacing
        return min_tick, -min_tick

    async def bitmap_range(self) -> tuple[int, int]:
        min_tick, max_tick = await self.tick_range()
        return (min_tick // 60 // 256), (max_tick // 60 // 256)

    async def get_initialized_ticks(self) -> list[int]:
        min_bitmap, max_bitmap = await self.bitmap_range()
        tick_spacing = await self.get_tick_spacing()
        bitmaps = await multicall.execute(
            *[
                self.tick_bitmap(primitives.int16(idx))
                for idx in range(min_bitmap, max_bitmap + 1)
            ]
        )
        initialized_ticks = []
        for idx, row in zip(range(min_bitmap, max_bitmap + 1), bitmaps):
            initialized_ticks.extend(tick_from_bitmap(idx, row, tick_spacing))
        return initialized_ticks

    async def active_tick(self) -> int:
        slot0 = await self.slot0().get()
        if not self._tick_spacing:
            self._tick_spacing = await self.tick_spacing().get()
        return slot0.tick // self._tick_spacing * self._tick_spacing

    async def liquidity_at_current_tick(self, tick: int) -> tuple[float, float]:
        tick_spacing, slot0, liquidity = cast(
            tuple[int, Slot0, int],
            await multicall.execute(
                self.tick_spacing(),
                self.slot0(),
                self.liquidity(),
            ),
        )
        bottom_tick = tick
        top_tick = tick + tick_spacing

        price = tick_to_price(slot0.tick)
        sa = tick_to_price(bottom_tick // 2)
        sb = tick_to_price(top_tick // 2)
        sqrt_price = math.sqrt(price)

        amount0 = liquidity * (sb - sqrt_price) / (sqrt_price * sb)
        amount1 = liquidity * (sqrt_price - sa)

        return (amount0, amount1)

    async def get_all_positions(self):
        events = []
        async for event in V3MintEvent.set_filter(addresses=[self.address]).backfill():
            events.append(event.event)
        return events

    def __repr__(self):
        return f"<Pool address={self.address}>"



from typing import Annotated

from eth_rpc.event import Event
from eth_rpc.types import Indexed, primitives
from eth_typing import HexAddress
from pydantic import BaseModel


class AddSharesType(BaseModel):
    wallet: Annotated[HexAddress, Indexed]
    amount: primitives.uint256


class RemoveSharesType(BaseModel):
    wallet: Annotated[HexAddress, Indexed]
    amount: primitives.uint256


class ClaimRewardType(BaseModel):
    wallet: Annotated[HexAddress, Indexed]


class DistributeRewardType(BaseModel):
    wallet: Annotated[HexAddress, Indexed]
    amount: primitives.uint256


class DepositRewardsType(BaseModel):
    wallet: Annotated[HexAddress, Indexed]
    amount: primitives.uint256


class DistributeRewardV2Type(BaseModel):
    wallet: Annotated[HexAddress, Indexed]
    token: Annotated[HexAddress, Indexed]
    amount: primitives.uint256


class DepositRewardsV2Type(BaseModel):
    wallet: Annotated[HexAddress, Indexed]
    token: Annotated[HexAddress, Indexed]
    amount: primitives.uint256


AddSharesEvent = Event[AddSharesType](name="AddShares")
RemoveSharesEvent = Event[RemoveSharesType](name="RemoveShares")
ClaimRewardEvent = Event[ClaimRewardType](name="ClaimReward")
DistributeRewardEvent = Event[DistributeRewardType](name="DistributeReward")
DepositRewardsEvent = Event[DepositRewardsType](name="DepositRewards")
DistributeRewardV2Event = Event[DistributeRewardV2Type](name="DistributeReward")
DepositRewardsV2Event = Event[DepositRewardsV2Type](name="DepositRewards")
